@@ -1,5 +1,5 @@
 package Positron::DataTemplate;
-our $VERSION = 'v0.0.3'; # VERSION
+our $VERSION = 'v0.0.4'; # VERSION
 
 =head1 NAME
 
@@ -7,7 +7,7 @@ Positron::DataTemplate - templating plain data to plain data
 
 =head1 VERSION
 
-version v0.0.3
+version v0.0.4
 
 =head1 SYNOPSIS
 
@@ -98,6 +98,12 @@ give some examples.
   [1, '. "/tmp/data.json"', 3] + '{ key: "value"}'
   -> [1, { key => 'value' }, 3]
 
+=head2 File wrapping (also requires L<JSON> and L<File::Slurp>)
+
+  [1, ': "/tmp/wrap.json"', { animal => 'dog' }, 3]
+  + '{ key: "value", contents: ":"}'
+  -> [1, { key => 'value', contents => { animal => 'dog' }, 3]
+
 =head2 Funtions on data
 
   [1, '^len', "abcde", 2] + { len => \&CORE::length }
@@ -128,48 +134,56 @@ sub process {
     # Returns (undef) in list context - is this correct?
     return undef unless defined $template;
     $env = Positron::Environment->new($env);
-    my @return = $self->_process($template, $env, '', 0);
-    # If called in scalar context, the caller "knows" that there will
-    # only be one element -> shortcut it.
-    return wantarray ? @return : $return[0];
-}
-
-sub _interpolate {
-    my ($value, $context, $interpolate) = @_;
-    return $value unless $interpolate;
-    if ($context eq 'array' and ref($value) eq 'ARRAY') {
-        return @$value;
-    } elsif ($context eq 'hash' and ref($value) eq 'HASH') {
-        return %$value;
+    my ($return, $interpolate) = $self->_process($template, $env);
+    # $return may be an interpolating construct,
+    # which depends on the context here.
+    if (wantarray and $interpolate and ref($return) eq 'ARRAY') {
+        return @$return;
     } else {
-        return $value;
+        return $return;
     }
 }
 
 sub _process {
-    my ($self, $template, $env, $context, $interpolate) = @_;
+    my ($self, $template, $env) = @_;
     if (not ref($template)) {
-        return $self->_process_text($template, $env, $context, $interpolate);
+        return $self->_process_text($template, $env);
     } elsif (ref($template) eq 'ARRAY') {
-        return $self->_process_array($template, $env, $context, $interpolate);
+        return $self->_process_array($template, $env);
     } elsif (ref($template) eq 'HASH') {
-        return $self->_process_hash($template, $env, $context, $interpolate);
+        return $self->_process_hash($template, $env);
     }
     return $template; # TODO: deep copy?
 }
 
 sub _process_text {
-    my ($self, $template, $env, $context, $interpolate) = @_;
+    my ($self, $template, $env) = @_;
+    return ($template, 0) unless $template; # undef, '', 0, or '0'
+    my $interpolate = 0;
     if ($template =~ m{ \A [&,] (-?) (.*) \z}xms) {
         if ($1) { $interpolate = 1; }
-        return _interpolate(Positron::Expression::evaluate($2, $env), $context, $interpolate);
+        my $expr = $2;
+        if ($expr eq ':') {
+            # Special case: internal wrap evaluation
+            my ($return, $i) = $self->_process($env->get(':'), $env);
+            $interpolate ||= $i;
+            return ($return, $interpolate);
+        } else {
+            return (Positron::Expression::evaluate($expr, $env), $interpolate);
+        }
     } elsif ($template =~ m{ \A \$ (.*) \z}xms) {
-        return "" . Positron::Expression::evaluate($1, $env);
+        return ("" . Positron::Expression::evaluate($1, $env), 0);
     } elsif ($template =~ m{ \A \x23 (\+?) }xms) {
-        return (wantarray and not $1) ? () : '';
-    } elsif ($template =~ m{ \A \. (-?) \s* (.*) }xms) {
-        if ($1) { $interpolate = 1; }
-        my $filename = Positron::Expression::evaluate($2, $env);
+        return ('', ($1 ? 0 : 1));
+    } elsif ($template =~ m{ \A ([.:]) (-?) \s* ([^\s-].*) }xms) {
+        my $filename_expr = $3;
+        if ($2) { $interpolate = 1; }
+        my $new_env = $env;
+        if ($1 eq ':') {
+            # A wrap in text context, explicitly unset ':'.
+            $new_env = Positron::Environment->new({ ':' => undef }, { parent => $env });
+        }
+        my $filename = Positron::Expression::evaluate($filename_expr, $new_env);
         require JSON;
         require File::Slurp;
         my $json = JSON->new();
@@ -180,15 +194,24 @@ sub _process_text {
             }
         }
         if ($file) {
-            my $result = $json->decode(File::Slurp::read_file($file));
-            return $self->_process($result, $env, $context, $interpolate);
+            my $result = $json->decode(scalar(File::Slurp::read_file($file)));
+            my ($return, $i) = $self->_process($result, $new_env);
+            $interpolate ||= $i;
+            return ($return, $interpolate);
         } else {
             croak "Can't find template '$filename' in " . join(':', @{$self->{include_paths}});
         }
-    } elsif ($template =~ m{ \A \^ (-?) \s* (.*) }xms) {
+    } elsif ($template =~ m{ \A \: (-?) \s* \z }xms) {
+        # wrap evaluation
         if ($1) { $interpolate = 1; }
+        my ($return, $i) = $self->_process($env->get(':'), $env);
+        $interpolate ||= $i;
+        return ($return, $interpolate);
+    } elsif ($template =~ m{ \A \^ (-?) \s* (.*) }xms) {
+        # Special non-list case, e.g. hash value (not key)
+        # cannot interpolate
         my $function = Positron::Expression::evaluate($2, $env);
-        return _interpolate($function->(), $context, $interpolate);
+        return (scalar($function->()), 0);
     } else {
         $template =~ s{
             \{ \$ ([^\}]*) \}
@@ -201,13 +224,14 @@ sub _process_text {
         }{
             $2 ? '' : $1 . $4;
         }xmseg;
-        return $template;
+        return ($template, 0);
     }
 }
 
 sub _process_array {
-    my ($self, $template, $env, $context, $interpolate) = @_;
-    return _interpolate([], $context, $interpolate) unless @$template;
+    my ($self, $template, $env) = @_;
+    my $interpolate = 0;
+    return ([], 0) unless @$template;
     my @elements = @$template;
     if ($elements[0] =~ m{ \A \@ (-?) (.*) \z}xms) {
         # list iteration
@@ -215,13 +239,20 @@ sub _process_array {
         my $clause = $2;
         shift @elements;
         my $result = [];
-        my $list = Positron::Expression::evaluate($clause, $env); # must be arrayref!
+        my $list = Positron::Expression::evaluate($clause, $env);
+        if (not ref($list) eq 'ARRAY') {
+            # If it's not a list, make it a one-element list.
+            # Useful for forcing interpolation via '[@- ""]' or aliasing (to be introduced)
+            $list = [$list];
+        }
         foreach my $el (@$list) {
             my $new_env = Positron::Environment->new( $el, { parent => $env } );
-            # Do not interpolate here, interpolate the result
-            push @$result, map $self->_process($_, $new_env, 'array', 0), @elements;
+            # evaluate rest of list as array,
+            my ($return, undef) = $self->_process_array(\@elements, $new_env);
+            # and flatten
+            push @$result, @$return;
         }
-        return _interpolate($result, $context, $interpolate);
+        return ($result, $interpolate);
     } elsif ($elements[0] =~ m{ \A \? (-?) (.*) \z}xms) {
         # conditional
         if ($1) { $interpolate = 1; }
@@ -230,23 +261,26 @@ sub _process_array {
         my $has_else = (@elements > 1) ? 1 : 0;
         my $cond = Positron::Expression::evaluate($clause, $env); # can be anything!
         # for Positron, empty lists and hashes are false!
-        # TODO: $cond = Positron::Expression::true($cond);
-        if (ref($cond) eq 'ARRAY' and not @$cond) { $cond = 0; }
-        if (ref($cond) eq 'HASH'  and not %$cond) { $cond = 0; }
+        $cond = Positron::Expression::true($cond);
         if (not $cond and not $has_else) {
-            # no else clause, return empty list on false
-            return ();
+            # no else clause, return empty on false
+            # (please interpolate!)
+            return ('', 1);
         }
         my $then = shift @elements;
         my $else = shift @elements;
         my $result = $cond ? $then : $else;
-        return $self->_process($result, $env, $context, $interpolate);
+        my ($return, $i) = $self->_process($result, $env);
+        $interpolate ||= $i;
+        return ($return, $interpolate);
     } else {
         my $return = [];
         # potential structural comments
         my $skip_next = 0;
         my $capturing_function = 0;
-        my $interpolate_next = 0;
+        my $capturing_wrap = 0;
+        my $capturing_wrap_interpolates = 0;
+        my $interpolate_next = 0; # actual count
         my $is_first_element = 1;
         foreach my $element (@elements) {
             if ($element =~ m{ \A // (-?) }xms) {
@@ -255,23 +289,85 @@ sub _process_array {
             } elsif ($element =~ m{ \A / (-?) }xms) {
                 if ($is_first_element and $1) { $interpolate = 1; }
                 $skip_next = 1;
-            } elsif ($element =~ m{ \A \^ (-?) \s* (.*) }xms) {
+            } elsif ($element =~ m{ \A \^ (-?) \s* ([^\s-].*) }xms) {
                 if ($is_first_element and $1) { $interpolate = 1; }
                 $capturing_function = Positron::Expression::evaluate($2, $env);
                 # do not push!
+            } elsif ($element =~ m{ \A \: (-?) \s* ([^\s-].*) }xms) {
+                $capturing_wrap_interpolates = $1 ? 1 : 0;
+                my $filename = Positron::Expression::evaluate($2, $env);
+                if (!$filename) {
+                    warn "# no filename in expression '$element'?";
+                }
+                require JSON;
+                require File::Slurp;
+                my $json = JSON->new();
+                my $file = undef;
+                foreach my $path (@{$self->{include_paths}}) {
+                    if (-f $path . $filename) {
+                        $file = $path . $filename; # TODO: platform-independent chaining
+                    }
+                }
+                if ($file) {
+                    my $contents = File::Slurp::read_file($file);
+                    $capturing_wrap = $json->decode($contents);
+                } else {
+                    croak "Can't find template '$filename' in " . join(':', @{$self->{include_paths}});
+                }
+                # do not push!
             } elsif ($skip_next) {
                 $skip_next = 0;
             } elsif ($capturing_function) {
                 # we have a capturing function waiting for input
-                my $arg = $self->_process($element, $env);
-                push @$return, $capturing_function->($arg);
+                my ($arg, $i) = $self->_process($element, $env);
+                # interpolate: could be ['@- ""', arg1, arg2]
+                if (ref($arg) eq 'ARRAY' and $i) {
+                    push @$return, $capturing_function->(@$arg);
+                } elsif (ref($arg) eq 'HASH' and $i) {
+                    push @$return, $capturing_function->(%$arg);
+                } else {
+                    push @$return, $capturing_function->($arg);
+                }
                 # no more waiting function
                 $capturing_function = 0;
+            } elsif ($capturing_wrap) {
+                # we have a capturing wrap file waiting for input
+                # Note: neither the wrap nor the element have been evaluated yet!
+                my $new_env = Positron::Environment->new({ ':' => $element }, { parent => $env });
+                my ($result, $i) = $self->_process($capturing_wrap, $new_env);
+                $i ||= $capturing_wrap_interpolates;
+                # interpolate: could be ['@- ""', arg1, arg2]
+                #              or [1, ':- file', 'contents', 2]
+                if (ref($result) eq 'ARRAY' and $i) {
+                    push @$return, @$result;
+                } elsif (ref($result) eq 'HASH' and $i) {
+                    push @$return, %$result;
+                } else {
+                    push @$return, $result;
+                }
+                # no more waiting wrap
+                $capturing_wrap = 0;
             } elsif ($element =~ m{ \A < }xms) {
-                $interpolate_next = 1;
+                $interpolate_next += 1; # actual count
             } else {
-                push @$return, $self->_process($element, $env, 'array', $interpolate_next);
+                my ($result, $interpolate_me) = $self->_process($element, $env);
+                my @results = ($result);
+                $interpolate_next += $interpolate_me;
+                while ($interpolate_next > 0 and @results) {
+                    if (ref($results[0]) eq 'ARRAY') {
+                        my $array = shift @results;
+                        unshift @results, @$array;
+                    } elsif (($results[0] // '') eq '') {
+                        # Note: the empty string, if it wants to interpolate, becomes the empty list
+                        #       i.e. just drop it.
+                        shift @results;
+                    } else {
+                        last; # conditions can't match any more
+                    }
+                    $interpolate_next--;
+                }
                 $interpolate_next = 0;
+                push @$return, @results;
             }
             $is_first_element = 0; # not anymore
         }
@@ -279,12 +375,24 @@ sub _process_array {
             # Oh no, a function waiting for args?
             push @$return, $capturing_function->();
         }
-        return _interpolate($return, $context, $interpolate);
+        if ($capturing_wrap) {
+            # Oh no, a wrap waiting for args?
+            my $new_env = Positron::Environment->new({ ':' => undef }, { parent => $env });
+            my ($result, $i) = $self->_process($capturing_wrap, $new_env);
+            if (ref($result) eq 'ARRAY' and $i) {
+                push @$return, @$result;
+            } elsif (ref($result) eq 'HASH' and $i) {
+                push @$return, %$result;
+            } else {
+                push @$return, $result;
+            }
+        }
+        return ($return, $interpolate);
     }
 }
 sub _process_hash {
-    my ($self, $template, $env, $context, $interpolate) = @_;
-    return _interpolate({}, $context, $interpolate) unless %$template;
+    my ($self, $template, $env) = @_;
+    return ({}, 0) unless %$template;
     my %result = ();
     my $hash_construct = undef;
     my $switch_construct = undef;
@@ -292,7 +400,7 @@ sub _process_hash {
         if ($key =~ m{ \A \% (.*) \z }xms) {
             $hash_construct = [$key, $1]; last;
         } elsif ($key =~ m{ \A \? (.*) \z }xms) {
-            # '?-': activate interpolate ?
+            # basically auto-interpolates
             $switch_construct = [$key, $1]; last;
         }
     }
@@ -301,22 +409,24 @@ sub _process_hash {
         croak "Error: result of expression '".$hash_construct->[1]."' must be hash" unless ref($e_content) eq 'HASH';
         while (my ($key, $value) = each %$e_content) {
             my $new_env = Positron::Environment->new( { key => $key, value => $value }, { parent => $env } );
-            my $t_content = $self->_process( $template->{$hash_construct->[0]}, $new_env);
+            my ($t_content, undef) = $self->_process( $template->{$hash_construct->[0]}, $new_env);
             croak "Error: content of % construct must be hash" unless ref($t_content) eq 'HASH';
-            # copy into result
+            # copy into result (automatically interpolates)
             foreach my $k (keys %$t_content) {
                 $result{$k} = $t_content->{$k};
             }
         }
     } elsif ($switch_construct) {
-        # '<': pass downwards ?
         my $e_content = Positron::Expression::evaluate($switch_construct->[1], $env); # The switch key
-        if (defined $e_content and exists $template->{$switch_construct->[0]}->{$e_content}) {
-            return $self->_process($template->{$switch_construct->[0]}->{$e_content}, $env);
+        # escape the '?' by adding another one!
+        my $qe_content = ( defined $e_content and $e_content =~m{ \A \?}xms ) ? "?$e_content" : $e_content;
+        if (defined $e_content and exists $template->{$switch_construct->[0]}->{$qe_content}) {
+            # We have no interpolation of our own, just pass the below up.
+            return $self->_process($template->{$switch_construct->[0]}->{$qe_content}, $env);
         } elsif (exists $template->{$switch_construct->[0]}->{'?'}) {
             return $self->_process($template->{$switch_construct->[0]}->{'?'}, $env);
         } else {
-            return ();
+            return ('', 1);
         }
     } else {
         # simple copy
@@ -341,8 +451,8 @@ sub _process_hash {
             my $value = $template->{$key};
             if ($key =~ m{ \A < }xms) {
                 # interpolate
-                my %values = $self->_process($value, $env, 'hash', 1);
-                %result = (%result, %values);
+                my ($values, $interpolate) = $self->_process($value, $env);
+                %result = (%result, %$values);
                 next;
             }
             if ($key =~ m{ \A / }xms) {
@@ -356,7 +466,7 @@ sub _process_hash {
             if ($key =~ m{ \A \^ \s* (.*)}xms) {
                 # consuming function call (interpolates)
                 my $func = Positron::Expression::evaluate($1, $env);
-                my $value_in = $self->_process($value, $env, '', 0);
+                my ($value_in, undef) = $self->_process($value, $env);
                 my $hash_out = $func->($value_in);
                 # interpolate
                 foreach my $k (keys %$hash_out) {
@@ -364,12 +474,39 @@ sub _process_hash {
                 }
                 next;
             }
-            $key = $self->_process($key, $env, '', 0);
-            $value = $self->_process($value, $env, '', 0);
+            if ($key =~ m{ \A : (-?) \s* (.+) }xms) {
+                # consuming wrap (interpolates in any case)
+                my $capturing_wrap;
+                my $filename = Positron::Expression::evaluate($2, $env);
+                require JSON;
+                require File::Slurp;
+                my $json = JSON->new();
+                my $file = undef;
+                foreach my $path (@{$self->{include_paths}}) {
+                    if (-f $path . $filename) {
+                        $file = $path . $filename; # TODO: platform-independent chaining
+                    }
+                }
+                if ($file) {
+                    my $contents = File::Slurp::read_file($file);
+                    $capturing_wrap = $json->decode($contents);
+                } else {
+                    croak "Can't find template '$filename' in " . join(':', @{$self->{include_paths}});
+                }
+                my $new_env = Positron::Environment->new({ ':' => $value }, { parent => $env });
+                my ($hash_out, undef) = $self->_process($capturing_wrap, $new_env);
+                # interpolate
+                foreach my $k (keys %$hash_out) {
+                    $result{$k} = $hash_out->{$k};
+                }
+                next;
+            }
+            ($key, undef) = $self->_process($key, $env);
+            ($value, undef) = $self->_process($value, $env);
             $result{$key} = $value;
         }
     }
-    return _interpolate(\%result, $context, $interpolate);
+    return (\%result, 0);
 }
 
 sub add_include_paths {
